# -*- coding: utf-8 -*-
"""
blueox.context
~~~~~~~~

This module provides the concept of 'Context' for collecting data that will
generate a log event.

:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""
import time
import os
import random
import struct
import threading
import functools
import logging

from . import utils
from . import network

log = logging.getLogger(__name__)

# This will be redefined by our configuration to tell us where to record stuff
# to.
_recorder_function = None


class Context(object):
    __slots__ = ["name", "data", "id", "_writable", "start_time",
                 "_sample_checks", "enabled"]

    def __init__(self, type_name, id=None, sample=None):
        """Create a new blueox logging context

        Arguments:
        type_name -- The Name for the event. Special prefix characters can
            control the name with respect to where this context fits into the
            heirarchy of parent requests. Examples:

            '.foo' - Will generate a name like '<parent name>.foo'
            '.foo.bar' - If the parent ends in '.foo', the final name will be '<parent name>.bar'
            '^.foo' - Will use the top-most context, generating '<top parent name>.foo'
            'top.foo.bar' - The name will be based on the longest matched
                parent context. If there is a parent context named 'top' and a
                parent context named 'top.foo', the new context will be named
                'top.foo.bar'

        id -- (optional) user-generated identifier

        sample -- (optional, tuple) of (sample parent, sample ratio). Examples:

            ('..', 0.25) will enable logging for 25% of the parent contexts.
            ('.', 0.25) will enable logging for 25% of the current context.
        """
        if type_name.startswith('.'):
            parent_ctx = current_context()

            if type_name.startswith('..'):
                raise ValueError("Invalid type name")

            clean_type_name = type_name[1:]
        elif type_name.startswith('^'):
            parent_ctx = top_context()

            if not type_name.startswith('^.'):
                raise ValueError("Invalid relative type name syntax")

            clean_type_name = type_name[2:]
        else:
            parent_ctx = find_closest_context(type_name)
            if parent_ctx is None:
                clean_type_name = type_name
            else:
                if parent_ctx.name == type_name:
                    # Previously we crashed with this being an invalid name,
                    # but we probably don't want to crash even though it's a
                    # highly unusual situation.
                    log.warning("Duplicate type name: %r", type_name)
                    clean_type_name = type_name
                elif type_name.startswith(parent_ctx.name):
                    # If the parent is a prefix of our current type name, we'll
                    # keep it as the parent, otherwise we're a separate branch
                    # of the context tree.
                    clean_type_name = type_name[len(parent_ctx.name) + 1:]
                else:
                    clean_type_name = type_name
                    parent_ctx = None

        if parent_ctx is None:
            self.name = clean_type_name
        else:
            parent_parts = sorted(enumerate(parent_ctx.name.split('.')),
                                  reverse=True)
            for ppart_ndx, ppart in parent_parts:
                if ppart == type_name.split('.')[1]:
                    self.name = '.'.join(parent_ctx.name.split('.')[:ppart_ndx]
                                         + type_name.split('.')[1:])
                    break
            else:
                self.name = '.'.join((parent_ctx.name, clean_type_name))

        self.data = {}
        self.start_time = time.time()
        self._sample_checks = {}

        if id is not None:
            self.id = id
        elif parent_ctx:
            self.id = parent_ctx.id
        else:
            # Generate an id if one wasn't provided and we don't have any parents
            # We're going to encode the time as the front 4 bytes so we have some order to the ids
            # that could prove useful later on by making sorting a little easier.
            self.id = (struct.pack(">L", time.time()) + os.urandom(12)).encode(
                'hex')

        if parent_ctx and not parent_ctx.enabled:
            self.enabled = False
        elif sample:
            sample_name, rate = sample

            if sample_name == type_name or sample_name == '.':
                sample_parent_ctx = self
            elif sample_name == '..':
                sample_parent_ctx = parent_ctx or self
            elif sample_name == '^':
                sample_parent_ctx = top_context() or self
            else:
                sample_parent_ctx = _get_context(sample_name)

            if sample_parent_ctx == self:
                self.enabled = bool(random.random() <= rate)
            else:
                self.enabled = sample_parent_ctx.sampled_for(type_name, rate)
        else:
            self.enabled = True

        self._writable = False

    @property
    def writable(self):
        """Indicates the contest is open and can be written to"""
        return self._writable

    def sampled_for(self, name, rate):
        if name not in self._sample_checks:
            self._sample_checks[name] = bool(random.random() <= rate)
        return self._sample_checks[name]

    def set(self, key, *args, **kwargs):
        if not self.writable:
            raise ValueError()

        if args and kwargs:
            raise ValueError()

        if len(args) > 1:
            utils.set_deep(self.data, key, args)
        elif args:
            utils.set_deep(self.data, key, args[0])
        elif kwargs:
            existing_value = utils.get_deep(self.data, key, {})
            existing_value.update(kwargs)
            utils.set_deep(self.data, key, existing_value)

    def append(self, key, value):
        if not self.writable:
            raise ValueError()

        existing_value = utils.get_deep(self.data, key, [])
        existing_value.append(value)
        if len(existing_value) == 1:
            utils.set_deep(self.data, key, existing_value)

    def add(self, key, value):
        if not self.writable:
            raise ValueError()

        existing_value = utils.get_deep(self.data, key, 0)
        utils.set_deep(self.data, key, existing_value + value)

    def to_dict(self):
        return {
            'id': self.id,
            'type': self.name,
            'host': os.uname()[1],
            'pid': os.getpid(),
            'start': self.start_time,
            'end': time.time(),
            'body': self.data
        }

    def start(self):
        _add_context(self)
        self._writable = True

    def stop(self):
        self._writable = False
        _remove_context(self)

    def done(self):
        self.stop()  # Just be sure
        if self.enabled and _recorder_function:
            _recorder_function(self)

        # Make sure we don't get any duplicate data
        # I would clear out all the data here, but that makes testing a little
        # more challenging.
        self.enabled = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.stop()
        self.done()


threadLocal = threading.local()


def init_contexts():
    if not getattr(threadLocal, 'init', None):
        threadLocal._contexts = []
        threadLocal._contexts_by_name = {}
        threadLocal.init = True


def clear_contexts():
    """Remove any currently active contexts.

    Does not 'close' any contexts, but just clears out any we may still have a
    handle on. This is useful in testing or anywhere else where you want to be
    absolutely sure you have a clean environment.
    """
    init_contexts()

    threadLocal._contexts_by_name.clear()
    del threadLocal._contexts[:]


def _add_context(context):
    init_contexts()

    if context.name in threadLocal._contexts_by_name:
        return
    threadLocal._contexts_by_name[context.name] = context
    threadLocal._contexts.append(context)


def _get_context(name):
    return threadLocal._contexts_by_name.get(str(name))


def _remove_context(context):
    init_contexts()
    try:
        del threadLocal._contexts_by_name[context.name]
    except KeyError:
        pass

    try:
        threadLocal._contexts.remove(context)
    except ValueError:
        pass


def current_context():
    init_contexts()
    try:
        return threadLocal._contexts[-1]
    except IndexError:
        return None


def top_context():
    """Return the outermost context"""
    init_contexts()
    try:
        return threadLocal._contexts[0]
    except IndexError:
        return None


def _calculate_match_length(a_parts, b_parts):
    length = 0
    for a, b in zip(a_parts, b_parts):
        if a != b:
            break
        length += 1

    return length


def find_closest_context(type_name):
    init_contexts()

    type_name_parts = type_name.split('.')
    if not type_name_parts[0]:
        raise ValueError(type_name)

    matched_ctx = None
    matched_ctx_len = 0
    for ctx in threadLocal._contexts:
        ctx_parts = ctx.name.split('.')

        shared_part_count = _calculate_match_length(type_name_parts, ctx_parts)

        if matched_ctx_len < shared_part_count:
            matched_ctx = ctx
            matched_ctx_len = shared_part_count

    return matched_ctx


def find_context(type_name):
    init_contexts()

    if type_name == '.':
        return current_context()
    elif type_name == '^':
        return top_context()
    elif type_name == '..':
        try:
            return threadLocal._contexts[-2]
        except IndexError:
            return None
    else:
        return _get_context(type_name)


def set(*args, **kwargs):
    context = current_context()
    if context:
        context.set(*args, **kwargs)


def append(*args, **kwargs):
    context = current_context()
    if context:
        context.append(*args, **kwargs)


def add(*args, **kwargs):
    context = current_context()
    if context:
        context.add(*args, **kwargs)


def context_wrap(type_name, sample=None):
    """Decorator for wrapping a function call with a context"""

    def wrapper(func):

        @functools.wraps(func)
        def inner(*args, **kwargs):
            with Context(type_name, sample=sample):
                return func(*args, **kwargs)

        return inner

    return wrapper



# -*- coding: utf-8 -*-
"""
blueox.ports
~~~~~~~~

This module provides utilities for generating connect strings.

BlueOx, thanks to it's 0mq roots has a somewhat complex relationship with
ports, which we'd like to abstract from the user as much as possible.

:copyright: (c) 2015 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""
import os

ENV_VAR_CONTROL_HOST = 'BLUEOX_CLIENT_HOST'
ENV_VAR_COLLECT_HOST = 'BLUEOX_HOST'

DEFAULT_HOST = '127.0.0.1'
DEFAULT_CONTROL_PORT = 3513
DEFAULT_COLLECT_PORT = 3514


def _default_host(host, default_host, default_port):
    """Build a default host string
    """
    if not host:
        host = default_host
    if ':' not in host:
        host = "{}:{}".format(host, default_port)

    return host


def default_control_host(host=None):
    default_host = os.environ.get(ENV_VAR_CONTROL_HOST, DEFAULT_HOST)
    return _default_host(host, default_host, DEFAULT_CONTROL_PORT)


def default_collect_host(host=None):
    default_host = os.environ.get(ENV_VAR_COLLECT_HOST, DEFAULT_HOST)
    return _default_host(host, default_host, DEFAULT_COLLECT_PORT)



# -*- coding: utf-8 -*-
"""
blueox.client
~~~~~~~~

This module provides utilities for writing client applications which connect or use blueox data.

:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""
import collections
import logging
import io
import sys
import itertools

import msgpack
import zmq

from . import ports
from . import store

log = logging.getLogger(__name__)


def default_host(host=None):
    """Build a default host string for clients

    This is specifically for the control port, so its NOT for use by loggers.
    We also respect environment variables BLUEOX_CLIENT_HOST and _PORT if
    command line options aren't your thing.
    """
    return ports.default_control_host(host)


def decode_stream(stream):
    """A generator which reads data out of the buffered file stream, unpacks and decodes the blueox events

    This is useful for parsing on disk log files generated by blueoxd
    """

    unpacker = msgpack.Unpacker()

    while True:
        try:
            data = stream.next()
        except StopIteration:
            break

        unpacker.feed(data)

        for msg in unpacker:
            yield msg


def retrieve_stream_host(context, control_host):
    poller = zmq.Poller()
    sock = context.socket(zmq.REQ)
    sock.connect("tcp://%s" % control_host)
    poller.register(sock, zmq.POLLIN)

    sock.send(msgpack.packb({'cmd': 'SOCK_STREAM'}))

    result = dict(poller.poll(5000))
    if sock in result:
        result = msgpack.unpackb(sock.recv())
        host, _ = control_host.split(':')
        return "%s:%d" % (host, result['port'])
    else:
        log.warning("Failed to connect to server")
        return None


def subscribe_stream(control_host, subscribe):
    context = zmq.Context()

    while True:
        stream_host = retrieve_stream_host(context, control_host)
        if stream_host is None:
            return

        sock = context.socket(zmq.SUB)

        prefix = False
        if subscribe:
            if subscribe.endswith('*'):
                prefix = True
                subscription = subscribe[:-1]
            else:
                subscription = subscribe
        else:
            subscription = ""

        sock.setsockopt(zmq.SUBSCRIBE, subscription)
        log.info("Connecting to %s" % (stream_host,))
        sock.connect("tcp://%s" % (stream_host,))

        # Now that we are connected, loop almost forever emiting events.
        # If we fail to receive any events within the specified timeout, we'll quit
        # and verify that we are connected to a valid stream.
        poller = zmq.Poller()
        poller.register(sock, zmq.POLLIN)
        while True:
            result = dict(poller.poll(5000))
            if sock not in result:
                break

            parts = sock.recv_multipart()
            if len(parts) == 2:
                channel, data = parts
                # If the client only want exact matches, we'll skip this guy.
                if not prefix and subscription and channel != subscription:
                    continue

                yield msgpack.unpackb(data)
            else:
                break


def stream_from_s3_store(bucket, type_name, start_dt, end_dt):
    log_files = store.find_log_files_in_s3(bucket, type_name, start_dt, end_dt)

    streams = []
    for lf in log_files:
        data_stream = lf.open(bucket)
        streams.append(decode_stream(data_stream))

    return itertools.chain(*streams)


def stdin_stream():
    stdin = io.open(sys.stdin.fileno(), buffering=0, mode='rb', closefd=False)
    stream = decode_stream(stdin)
    return stream


class Grouper(object):
    """Utility for grouping events and sub-events together.
    
    Events fed into a Grouper are joined by their common 'id'. Encountering the
    parent event type will trigger emitting a list of all events and sub events
    for that single id. 

    This assumes that the parent event will be the last encountered.

    So for example, you might do something like:

        stream = blueox.client.decode_stream(stdin)
        for event_group in client.Grouper(stream):
            ... do some processing of the event group ...

    """

    def __init__(self, stream, max_size=1000):
        self.max_size = max_size
        self.stream = stream
        self.dict = collections.OrderedDict()

    @property
    def size(self):
        return len(self.dict)

    def __iter__(self):
        for event in self.stream:

            while self.size > self.max_size:
                self.dict.popitem(last=False)

            try:
                self.dict[event['id']].append(event)
            except KeyError:
                self.dict[event['id']] = [event]

            if '.' not in event['type']:
                yield self.dict.pop(event['id'])

        raise StopIteration



# -*- coding: utf-8 -*-
"""
blueox.network
~~~~~~~~

This module provides our interface into ZeroMQ

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.

"""
import logging
import threading
import struct
import atexit

import zmq
import msgpack

from . import utils

log = logging.getLogger(__name__)

# We want to limit how many messages we'll hold in memory so if our oxd is
# unavailable, we don't just run out of memory.  I based this value on rough
# value of rather large 3k sized messages, and how many we can fit in 10 megs.
MAX_QUEUED_MESSAGES = 3500

# If we have pending outgoing messages, this is how long we'll wait after
# being told to exit.
LINGER_SHUTDOWN_MSECS = 2000

META_STRUCT_FMT = "!Bd64p64p"

# We're going to include a version byte in our meta struct for future
# upgrading.  This is very quickly getting into the kind of bit packing I
# wanted to avoid for a logging infrastructure, but the performance gain is
# hard to ignore.
META_STRUCT_VERSION = 0x3


def check_meta_version(meta):
    value, = struct.unpack(">B", meta[0])
    if value != META_STRUCT_VERSION:
        raise ValueError(value)


threadLocal = threading.local()

# Context can be shared between threads
_zmq_context = None
_connect_str = None


def init(host, port):
    global _zmq_context
    global _connect_str

    _zmq_context = zmq.Context()
    _connect_str = "tcp://%s:%d" % (host, port)


def _thread_connect():
    if _zmq_context and not getattr(threadLocal, 'zmq_socket', None):
        threadLocal.zmq_socket = _zmq_context.socket(zmq.PUSH)
        threadLocal.zmq_socket.hwm = MAX_QUEUED_MESSAGES
        threadLocal.zmq_socket.linger = LINGER_SHUTDOWN_MSECS

        threadLocal.zmq_socket.connect(_connect_str)


def _serialize_context(context):
    # Our sending format is made up of two messages. The first has a
    # quick to unpack set of meta data that our collector is going to
    # use for routing and stats. This is much faster than having the
    # collector decode the whole event. We're just going to use python
    # struct module to make a quick and dirty data structure
    context_dict = context.to_dict()
    for key in ('host', 'type'):
        if len(context_dict.get(key, "")) > 64:
            raise ValueError("Value too long: %r" % key)

    meta_data = struct.pack(META_STRUCT_FMT, META_STRUCT_VERSION,
                            context_dict['end'], context_dict['host'],
                            context_dict['type'])

    try:
        context_data = msgpack.packb(context_dict)
    except TypeError:
        try:
            # If we fail to serialize our context, we can try again with an
            # enhanced packer (it's slower though)
            context_data = msgpack.packb(context_dict,
                                         default=utils.msgpack_encode_default)
        except TypeError:
            log.exception("Serialization failure (not fatal, dropping data)")

            # One last try after dropping the body
            context_dict['body'] = None
            context_data = msgpack.packb(context_dict)

    return meta_data, context_data


def send(context):
    global _zmq_context
    _thread_connect()

    try:
        meta_data, context_data = _serialize_context(context)
    except Exception:
        log.exception("Failed to serialize context")
        return

    if _zmq_context and threadLocal.zmq_socket is not None:
        try:
            log.debug("Sending msg")
            threadLocal.zmq_socket.send_multipart(
                (meta_data, context_data), zmq.NOBLOCK)
        except zmq.ZMQError, e:
            log.exception("Failed sending blueox event, buffer full?")
    else:
        log.info("Skipping sending event %s", context.name)


def close():
    global _zmq_context

    if getattr(threadLocal, 'zmq_socket', None):
        threadLocal.zmq_socket.close()
        threadLocal.zmq_socket = None

    _zmq_context = None


atexit.register(close)



# -*- coding: utf-8 -*-
"""
blueox.store
~~~~~~~~

This module provides utility functions for managing log files as created by
oxd.

:copyright: (c) 2015 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""
import logging
import datetime
import os
import re
import collections
import io
import bz2

try:
    import boto
except ImportError:
    boto = None
else:
    from boto.s3.connection import OrdinaryCallingFormat

from . import errors

log = logging.getLogger(__name__)

DATE_FORMATS = ["%Y%m%d", "%Y%m%d %H:%M"]


class InvalidDateError(errors.Error):
    pass


def parse_date_range_argument(value):
    for fmt in DATE_FORMATS:
        try:
            return datetime.datetime.strptime(value, fmt)
        except ValueError:
            pass

    raise InvalidDateError()


class LogFile(object):
    """Represents a log file

    Primarily this is useful for generation and interpretation of the bits of
    data that form the log's filename and path. Sub-classes implement
    functionality to actually manipulate log files.
    """

    def __init__(self, type_name, host=None, dt=None, date=None, bzip=False):
        self.type_name = type_name
        self.host = host
        self.bzip = bzip
        if (dt, date) == (None, None):
            raise ValueError("Needs a date")

        # LogFile's may be based on just a date, or a full datetime object.
        # This has to do with whether rotation (and thus the naming of the log
        # files) will result in hourly offsets or just a single log for each
        # (day, type) combination.
        self.dt = dt
        self.date = date or dt.date()

    @property
    def sort_dt(self):
        # Log files may not represent an actual datetime, but sometimes we need
        # to sort them like they are, preferrably incorporating the fact that
        # users may switch from daily rotation (the default) to hourly.
        if self.dt:
            return self.dt
        else:
            return datetime.datetime(
                self.date.year, self.date.month, self.date.day)

    @property
    def file_name(self):
        if self.dt:
            date_name_str = self.dt.strftime('%Y%m%d%H')
        else:
            date_name_str = self.date.strftime('%Y%m%d')

        bzip_str = ""
        if self.bzip:
            bzip_str = ".bz2"

        host_str = ""
        if self.host:
            host_str = "-{}".format(self.host)

        return "{type}-{date_name}{host_str}.log{bzip}".format(
            type=self.type_name,
            host_str=host_str,
            date_name=date_name_str,
            bzip=bzip_str)

    @property
    def file_path(self):
        date_str = self.date.strftime('%Y%m%d')
        return "{date}/{filename}".format(
            date=date_str,
            filename=self.file_name)

    @classmethod
    def from_filename(cls, filename):
        basename = os.path.basename(filename)
        match = re.match(
            r"^(?P<stream>.+)"  # stream name like "nginx-error"
            r"\-(?P<date>\d{8,10})"  # date like 20140229 or 2014022910
            r"\-?(?P<host>.+)?"  # optional server name
            r"\.log"
            r"(?P<zip>\.bz2|\.gz)?$", basename)

        if match is None:
            raise ValueError(basename)

        match_info = match.groupdict()

        log_dt = log_date = None
        if len(match_info['date']) == len('yyyymmddHH'):
            log_dt = datetime.datetime.strptime(match_info["date"], '%Y%m%d%H')
        else:
            log_date = datetime.datetime.strptime(match_info["date"],
                                                  '%Y%m%d').date()

        return cls(
            match_info["stream"],
            host=match_info.get('host'),
            dt=log_dt,
            date=log_date,
            bzip=bool(match_info['zip']))


class S3LogFile(LogFile):

    def s3_key(self, bucket):
        return boto.s3.key.Key(bucket, name=self.file_path)

    def open(self, bucket):
        """Create a iterable stream of data from the log file.

        Automatically handles bzip decoding
        """

        def stream():
            if self.bzip:
                decompressor = bz2.BZ2Decompressor()
            else:
                decompressor = None

            for data in self.s3_key(bucket):
                if decompressor:
                    r = decompressor.decompress(data)
                else:
                    r = data

                if r is not None:
                    yield r

        return stream()

    @classmethod
    def from_s3_key(cls, key):
        return cls.from_filename(key.name)


class LocalLogFile(LogFile):

    def get_local_file_path(self, log_path):
        return os.path.join(log_path, self.file_path)

    def open(self, log_path):
        """Create a iterable stream of data from the log file.

        Automatically handles bzip decoding
        """

        def stream():
            if self.bzip:
                decompressor = bz2.BZ2Decompressor()
            else:
                decompressor = None

            with io.open(self.get_local_file_path(log_path), "rb") as f:
                for data in f:
                    if decompressor:
                        r = decompressor.decompress(data)
                    else:
                        r = data

                    if r is not None:
                        yield r

        return stream()

    def build_remote(self, host):
        return S3LogFile(
            self.type_name,
            host=host,
            dt=self.dt,
            date=self.date,
            bzip=self.bzip)


def list_log_files(log_path):
    """Find and parse all the log files in the specified log path"""
    log_files = []
    for dirpath, dirnames, filenames in os.walk(log_path):
        for filename in filenames:
            full_path = os.path.join(dirpath, filename)

            try:
                log_file = LocalLogFile.from_filename(filename)
            except ValueError:
                log.warning("Not a blueox log file: %s", full_path)
                continue
            log_files.append(log_file)

    return log_files


def filter_log_files_for_active(log_files):
    """Filter our list of log files to remove those we expect might be active."""
    out_log_files = []

    files_by_type = collections.defaultdict(list)
    for lf in log_files:
        files_by_type[lf.type_name].append(lf)

    for type_files in files_by_type.values():
        type_files.sort(key=lambda f: f.sort_dt)

        # We assume only the last log file in the list can be possibly be in
        # use.
        last_lf = type_files.pop()

        out_log_files += type_files

        # If that last log file is old, then it's probably not being used either.
        # We add a buffer of an hour just to make sure everything has rotated
        # away safely when this is run close to midnight.
        cutoff_date = (datetime.datetime.utcnow() - datetime.timedelta(hours=1)
                      ).date()
        if last_lf.date < cutoff_date:
            out_log_files.append(last_lf)

    return out_log_files


def filter_log_files_for_zipping(log_files):
    """Identify unzipped log files that are approporate for zipping.

    Each unique log type found should have the most recent log file unzipped
    as it's probably still in use.
    """
    out_files = []
    for lf in filter_log_files_for_active(log_files):
        if lf.bzip:
            continue

        out_files.append(lf)

    return out_files


def filter_log_files_for_uploading(log_files, zipped_only):
    """Filter out log files that we shouldn't upload

    specify zipped_only if we should only bother to upload zipped log files
    """
    out_files = []
    for lf in filter_log_files_for_active(log_files):
        if zipped_only and not lf.bzip:
            continue

        out_files.append(lf)

    return out_files


def zip_log_file(log_file, log_path):
    orig_path = log_file.get_local_file_path(log_path)

    log_file.bzip = True

    zip_path = log_file.get_local_file_path(log_path)

    # It's hard to believe, but this appears in testing to be just as fast as
    # spawning a bzip2 process.
    zip_file = bz2.BZ2File(zip_path, 'w', io.DEFAULT_BUFFER_SIZE)

    with io.open(orig_path, "rb") as fp:
        for data in fp:
            zip_file.write(data)
    zip_file.close()
    os.unlink(orig_path)


def s3_prefix_for_date_and_type(date, type_name):
    date_str = date.strftime('%Y%m%d')
    return "{}/{}-".format(date_str, type_name)


def inclusive_date_range(start_dt, end_dt):
    start_date = start_dt.date()
    end_date = end_dt.date()

    current_date = start_date
    while current_date <= end_date:
        yield current_date
        current_date += datetime.timedelta(days=1)


def find_log_files_in_s3(bucket, type_name, start_dt, end_dt):
    prefixes = []
    for dt in inclusive_date_range(start_dt, end_dt):
        prefixes.append(s3_prefix_for_date_and_type(dt, type_name))

    log_files = []
    for pf in prefixes:
        for key in bucket.list(pf):
            try:
                log_files.append(S3LogFile.from_s3_key(key))
            except ValueError as e:
                log.warning("S3 key %r not a log file: %r", key, e)
                continue

    out_log_files = []
    for lf in log_files:
        assert lf.type_name == type_name

        if lf.dt is None:
            # For log files that cover an entire day, we'll include it even if
            # the query was for a specific time range.
            out_log_files.append(lf)

        elif lf.dt >= start_dt and lf.dt <= end_dt:
            out_log_files.append(lf)

    return out_log_files


def find_log_files_in_path(log_path, type_name, start_dt, end_dt):
    log_files = list_log_files(log_path)
    out_log_files = []

    for lf in log_files:
        if lf.dt is None:
            if lf.date < start_dt.date() or lf.date > end_dt.date():
                continue
        else:
            if lf.dt < start_dt or lf.dt > end_dt:
                continue

        out_log_files.append(lf)

    return out_log_files


def open_bucket(bucket_name):
    region_name = os.environ.get('AWS_DEFAULT_REGION', 'us-east-1')

    # We need to specify calling_format to get around bugs in having a '.' in
    # the bucket name
    conn = boto.s3.connect_to_region(region_name,
                                     calling_format=OrdinaryCallingFormat())

    return conn.get_bucket(bucket_name)



# -*- coding: utf-8 -*-
"""
blueox.tornado
~~~~~~~~

This module provides hooks for using blueox with the Tornado async web server.
Making blueox useful inside tornado is a challenge since you'll likely want a
blueox context per request, but multiple requests can be going on at once inside
tornado.

:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""
import functools
import logging
import traceback
import types
import sys
import time

log = logging.getLogger(__name__)

import tornado.web
import tornado.gen
import tornado.httpclient
import tornado.simple_httpclient
import tornado.stack_context

import blueox


def _gen_wrapper(ctx, generator):
    """Generator Wrapper that starts/stops our context
    """
    try:
        yielded = None
        while True:
            ctx.start()

            value = generator.send(yielded)

            ctx.stop()

            try:
                yielded = yield value
            except Exception:
                ctx.start()
                value = generator.throw(*sys.exc_info())
                ctx.stop()

                yielded = yield value

    except (tornado.gen.Return, StopIteration):
        ctx.done()
        raise


def coroutine(func):
    """Replacement for tornado.gen.coroutine that manages a blueox context

    The difficulty with managing global blueox contexts in an async environment
    is contexts will need to start and stop depending on what steps of a
    coroutine are running. This decorator wraps the default coroutine decorator
    allowing us to stop and restore the context whenever this coroutine runs.

    If you don't use this wrapper, unrelated contexts may be grouped together!
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            ctx = args[0].blueox_ctx
        except (AttributeError, IndexError):
            ctx = None

        # Remember, not every coroutine wrapped method will return a generator,
        # so we have to manage context switching in multiple places.
        if ctx is not None:
            ctx.start()

        result = func(*args, **kwargs)

        if ctx is not None:
            ctx.stop()

        if isinstance(result, types.GeneratorType):
            return _gen_wrapper(ctx, result)

        return result

    real_coroutine = tornado.gen.coroutine
    return real_coroutine(wrapper)


class BlueOxRequestHandlerMixin(object):
    """Include in a RequestHandler to get a blueox context for each request

    """
    blueox_name = "request"

    def prepare(self):
        self.blueox_ctx = blueox.Context(self.blueox_name)
        self.blueox_ctx.start()
        super(BlueOxRequestHandlerMixin, self).prepare()

    def on_finish(self):
        super(BlueOxRequestHandlerMixin, self).on_finish()
        self.blueox_ctx.done()
        self.blueox_ctx = None


class SampleRequestHandler(BlueOxRequestHandlerMixin,
                           tornado.web.RequestHandler):
    """Sample base request handler that provides basic information about the request.
    """

    def prepare(self):
        super(SampleRequestHandler, self).prepare()
        blueox.set('headers', self.request.headers)
        blueox.set('method', self.request.method)
        blueox.set('uri', self.request.uri)

    def write_error(self, status_code, **kwargs):
        if 'exc_info' in kwargs:
            blueox.set('exception',
                       ''.join(traceback.format_exception(*kwargs["exc_info"])))

        return super(SampleRequestHandler, self).write_error(status_code,
                                                             **kwargs)

    def write(self, chunk):
        blueox.add('response_size', len(chunk))
        return super(SampleRequestHandler, self).write(chunk)

    def on_finish(self):
        blueox.set('response_status_code', self._status_code)
        super(SampleRequestHandler, self).on_finish()


class AsyncHTTPClient(tornado.simple_httpclient.SimpleAsyncHTTPClient):

    def __init__(self, *args, **kwargs):
        self.blueox_name = '.httpclient'
        return super(AsyncHTTPClient, self).__init__(*args, **kwargs)

    def fetch(self, request, callback=None, **kwargs):
        start_time = time.time()

        if isinstance(request, basestring):
            request = tornado.httpclient.HTTPRequest(url=request, **kwargs)

        ctx = blueox.Context(self.blueox_name)
        ctx.start()
        ctx.set('request.uri', request.url)
        ctx.set('request.method', request.method)
        ctx.set('request.size', len(request.body) if request.body else 0)

        ctx.stop()

        # I'd love to use the future to handle the completion step, BUT, we
        # need this to happen first. If the caller has provided a callback, we don't want them
        # to get called before we do. Rather than poke into the internal datastructures, we'll just 
        # handle the callback explicitly

        def complete_context(response):
            ctx.start()

            ctx.set('response.code', response.code)
            ctx.set('response.size', len(response.body) if response.body else 0)

            ctx.done()

        if callback is None:

            def fetch_complete(future):
                # This error handling is just copied from tornado.httpclient as
                # we need to record a real HTTPError. httpclient might do the same thing
                # again if needs to deal with the caller's callbacks.
                exc = future.exception()
                if isinstance(
                    exc,
                    tornado.httpclient.HTTPError) and exc.response is not None:
                    response = exc.response
                elif exc is not None:
                    response = tornado.httpclient.HTTPResponse(
                        request, 599,
                        error=exc,
                        request_time=time.time() - start_time)
                else:
                    response = future.result()

                complete_context(response)

            future = super(AsyncHTTPClient, self).fetch(request)
            future.add_done_callback(fetch_complete)
        else:

            def callback_wrapper(response):
                complete_context(response)
                callback(response)

            future = super(AsyncHTTPClient,
                           self).fetch(request,
                                       callback=callback_wrapper)

        return future



import blueox
import traceback
import sys
from flask import request, got_request_exception


class BlueOxMiddleware(object):
    """BlueOx Application Middleware for request/response objects

    BlueOx middleware has 3 callbacks registered with flask:
        before_request:
            Initialize a blueox context and populate with metadata about the
            request such as header information, request url, and client ip.
        after_request:
            Add more metadata about the response back to the client then chean
            up and flush the corresponding context created during
            before_request.
        handle_exception:
            When an exception is fired, flask will dispatch to this function to
            log it to blueox.
    """

    def __init__(self, app):
        self.app = app

        if 'BLUEOX_HOST' in app.config:
            self.blueox_host = app.config['BLUEOX_HOST']
            if self.blueox_host:
                blueox.default_configure(self.blueox_host)
        else:
            blueox.default_configure()

        self.app.before_request(self.before_request)
        self.app.after_request(self.after_request)

        got_request_exception.connect(self.handle_exception, self.app)

    def before_request(self, *args, **kwargs):
        request.blueox = blueox.Context(
            ".".join((self.app.config.get('BLUEOX_NAME', ''), 'request')))

        blueox.set('method', request.method)
        blueox.set('path', request.path)

        headers = {}
        for k, v in request.environ.iteritems():
            if (
                k.startswith('HTTP_') or k in
                ('CONTENT_LENGTH', 'CONTENT_TYPE')):
                headers[k] = v

        blueox.set('headers', headers)
        blueox.set('url', request.url)
        blueox.set('client_ip', request.environ.get('REMOTE_ADDR'))

    def after_request(self, response):
        if not hasattr(request, 'blueox'):
            return

        if hasattr(request, 'user'):
            blueox.set('user', request.user.id)

        for key in ('version', 'revision'):
            if hasattr(request, key):
                blueox.set(key, getattr(request, key))

        if response is not None:
            blueox.set('response_status_code', response.status_code)

            if not response.is_streamed:
                blueox.set('response_size', response.content_length)

        request.blueox.done()

        return response

    def handle_exception(self, *args, **kwargs):
        blueox.set(
            'exception', ''.join(traceback.format_exception(*sys.exc_info())))



# -*- coding: utf-8 -*-
"""
blueox.timer
~~~~~~~~

This module has a timer context manager for easily tracking wall-clock time for some execution
:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""
import time

from . import context


class Timer(object):

    def __init__(self, context, key):
        self.context = context
        self.key = key
        self.start_time = time.time()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.context.set(self.key, time.time() - self.start_time)


def timeit(key):
    return Timer(context.current_context(), key)



# -*- coding: utf-8 -*-
"""
blueox.utils
~~~~~~~~

This module provides utility functions that are used within BlueOx

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.

"""
import datetime
import decimal
import time


class ParsedKey(object):

    def __init__(self, value):
        self.elems = []
        if '.' in value:
            for part in value.split('.'):
                try:
                    self.elems.append(int(part))
                except ValueError:
                    self.elems.append(part)
        else:
            self.elems.append(value)

    def __getitem__(self, key):
        return self.elems[key]

    def __getslice__(self, i, j, seq=None):
        usable_elems = self.elems[i:j:seq]
        if usable_elems:
            return ParsedKey('.'.join((str(v) for v in usable_elems)))
        else:
            return []

    def __len__(self):
        return len(self.elems)

    def __str__(self):
        return '.'.join((str(v) for v in self.elems))


def parse_key(key):
    return ParsedKey(key)


def get_deep(target, key, default=None):
    value = target
    for elem in parse_key(key):
        try:
            value = value[elem]
        except (KeyError, IndexError):
            return default

    return value


def set_deep(target, key, value):
    iter_value = target
    p_key = parse_key(key)
    for elem in p_key[:-1]:
        if isinstance(elem, int):
            raise ValueError(elem)

        iter_value = iter_value.setdefault(elem, {})

    iter_value[p_key[-1]] = value


def msgpack_encode_default(obj):
    """Extra encodings for python types into msgpack

    These are attempted if our normal serialization fails.
    """
    if isinstance(obj, decimal.Decimal):
        return str(obj)
    if isinstance(obj, datetime.datetime):
        return time.mktime(obj.utctimetuple())
    if isinstance(obj, datetime.date):
        return obj.strftime("%Y-%m-%d")

    raise TypeError("Unknown type: %r" % (obj,))



# -*- coding: utf-8 -*-
"""
blueox.logger
~~~~~~~~

This module provides integration with blueox and standard python logging module.
:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""
import logging
import traceback

from .context import Context


class LogHandler(logging.Handler):
    """Handler to provide log events as blueox events.

    Records standard fields such as logger name, level the message and if an
    exception was provided, the string formatted exception.

    The type name, if not specified will be something like '<my parent context>.log'
    """

    def __init__(self, type_name=None):
        super(LogHandler, self).__init__()

        self.type_name = type_name

    def emit(self, record):
        # We don't want any endless recursion
        if record.name.startswith('blueox'):
            return

        with Context(self.type_name or '.log') as c:
            c.set('name', record.name)
            c.set('level', record.levelname)
            c.set('msg', record.getMessage())
            if record.exc_info:
                c.set('exception',
                      ''.join(traceback.format_exception(*record.exc_info)))



"""Hooks for gathering celery task data into blueox.

Importing this module will register signal handlers into Celery worker's runtime.

We also will track creation of tasks on the client side.
"""
import traceback

import blueox
from blueox.context import current_context
from celery import states
from celery import signals
from django.conf import settings

# If dealer is installed, use it to get our revision
try:
    import dealer.auto
    revision = dealer.auto.auto.revision
except ImportError:
    revision = None

if hasattr(signals, 'after_task_publish'):

    @signals.after_task_publish.connect
    def on_task_sent(sender=None, body=None, **kwargs):
        with blueox.Context('.celery.task_sent'):
            blueox.set('task_id', body['id'])
            blueox.set('task', str(body['task']))
            blueox.set('eta', body['eta'])

else:

    @signals.task_sent.connect
    def on_task_sent(**kwargs):
        with blueox.Context('.celery.task_sent'):
            # Arguments for this signal are different than the worker signals. Sometimes
            # they are even different than what the documentation says. See also
            # https://github.com/celery/celery/issues/1606
            blueox.set('task_id', kwargs.get('task_id', kwargs.get('id')))
            blueox.set('task', str(kwargs['task']))
            blueox.set('eta', kwargs['eta'])


@signals.worker_process_init.connect
def on_worker_process_init(**kwargs):
    if hasattr(settings, 'BLUEOX_HOST'):
        if settings.BLUEOX_HOST:
            blueox.default_configure(settings.BLUEOX_HOST)
        else:
            blueox.configure(None, None)
    else:
        blueox.default_configure()


@signals.worker_shutdown.connect
def on_worker_shutdown(**kwargs):
    blueox.shutdown()


@signals.task_prerun.connect
def on_task_prerun(**kwargs):
    ctx = blueox.Context(".".join(
        (getattr(settings, 'BLUEOX_NAME', ''), 'celery', 'task')))
    ctx.start()

    ctx.set('task', kwargs['task'].name)
    ctx.set('retries', kwargs['task'].request.retries)
    ctx.set('expires', kwargs['task'].request.expires)
    ctx.set('delivery_info', kwargs['task'].request.delivery_info)
    ctx.set('task_id', str(kwargs['task_id']))
    ctx.set('args', kwargs['args'])
    ctx.set('kwargs', kwargs['kwargs'])

    if revision:
        ctx.set('version', revision)


@signals.task_failure.connect
def on_task_failure(**kwargs):
    blueox.set('exception', "".join(
        traceback.format_exception(*kwargs['einfo'].exc_info)))


@signals.task_retry.connect
def on_task_retry(**kwargs):
    blueox.set('result_state', states.RETRY)

    blueox.set('exception', "".join(
        traceback.format_exception(*kwargs['einfo'].exc_info)))

    # Retry path doesn't call 'postrun'. I think in celery-speak, it's the same
    # task, but we want to track attempts.
    ctx = current_context()
    if ctx:
        ctx.done()


@signals.task_postrun.connect
def on_task_postrun(**kwargs):
    blueox.set('result_state', str(kwargs['state']))

    ctx = current_context()
    if ctx:
        ctx.done()



import sys
import traceback
import logging

import blueox

from django.conf import settings


class Middleware(object):

    def __init__(self):
        if hasattr(settings, 'BLUEOX_HOST'):
            if settings.BLUEOX_HOST:
                blueox.default_configure(settings.BLUEOX_HOST)
            else:
                blueox.configure(None, None)
        else:
            blueox.default_configure()

    def process_request(self, request):
        request.blueox = blueox.Context(".".join(
            (getattr(settings, 'BLUEOX_NAME', ''), 'request')))
        request.blueox.start()

        blueox.set('method', request.method)
        blueox.set('path', request.path)

        headers = {}
        for k, v in request.META.iteritems():
            if k.startswith('HTTP_') or k in ('CONTENT_LENGTH', 'CONTENT_TYPE'):
                headers[k] = v
        blueox.set('headers', headers)

        blueox.set('uri', request.build_absolute_uri())
        blueox.set('client_ip', request.META['REMOTE_ADDR'])

        return None

    def process_response(self, request, response):
        # process_request() is not guaranteed to be called
        if not hasattr(request, 'blueox'):
            return response

        # We collect some additional data in the response, just to ensure
        # middleware ordering doesn't matter.
        if hasattr(request, 'user'):
            blueox.set('user', request.user.id)

        for key in ('version', 'revision'):
            if hasattr(request, key):
                blueox.set(key, getattr(request, key))

        # Other middleware may have blocked our response.
        if response is not None:

            blueox.set('response_status_code', response.status_code)

            if not response.streaming:
                blueox.set('response_size', len(response.content))

            headers = {}
            for k, v in response.items():
                headers[k] = v

            blueox.set('response_headers', headers)

        request.blueox.done()

        return response

    def process_exception(self, request, exception):
        blueox.set('exception',
                   ''.join(traceback.format_exception(*sys.exc_info())))
        return None



# -*- coding: utf-8 -*-
"""
This module contains the set of Blue Ox's exceptions

:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""


class Error(Exception):
    """This is an ambiguous error that occured."""
    pass



# -*- coding: utf-8 -*-
"""
blueox
~~~~~~~~

:copyright: (c) 2012 by Rhett Garber
:license: ISC, see LICENSE for more details.

"""

__title__ = 'blueox'
__version__ = '0.11.0'
__author__ = 'Rhett Garber'
__author_email__ = 'rhettg@gmail.com'
__license__ = 'ISC'
__copyright__ = 'Copyright 2015 Rhett Garber'
__description__ = 'A library for python-based application logging and data collection'
__url__ = 'https://github.com/rhettg/BlueOx'

import logging
import os

from . import utils
from . import network
from . import ports
from .context import (
    Context, set, append, add, context_wrap, current_context, find_context,
    clear_contexts)
from . import context as _context_mod
from .errors import Error
from .logger import LogHandler
from .timer import timeit

log = logging.getLogger(__name__)


def configure(host, port, recorder=None):
    """Initialize blueox

    This instructs the blueox system where to send it's logging data. If blueox is not configured, log data will
    be silently dropped.

    Currently we support logging through the network (and the configured host and port) to a blueoxd instances, or
    to the specified recorder function
    """
    if recorder:
        _context_mod._recorder_function = recorder
    elif host and port:
        network.init(host, port)
        _context_mod._recorder_function = network.send
    else:
        log.info("Empty blueox configuration")
        _context_mod._recorder_function = None


def default_configure(host=None):
    """Configure BlueOx based on defaults

    Accepts a connection string override in the form `localhost:3514`. Respects
    environment variable BLUEOX_HOST
    """
    host = ports.default_collect_host(host)
    hostname, port = host.split(':')

    try:
        int_port = int(port)
    except ValueError:
        raise Error("Invalid value for port")

    configure(hostname, int_port)


def shutdown():
    network.close()
